@@ -1,8 +1,9 @@
+from __future__ import annotations
+
 from tornado.ioloop import IOLoop
 
 from distributed.dashboard.components.worker import (
     counters_doc,
-    crossfilter_doc,
     profile_doc,
     profile_server_doc,
     status_doc,
@@ -10,9 +11,7 @@ from distributed.dashboard.components.worker import (
 )
 from distributed.dashboard.core import BokehApplication
 
-template_variables = {
-    "pages": ["status", "system", "profile", "crossfilter", "profile-server"]
-}
+template_variables = {"pages": ["status", "system", "profile", "profile-server"]}
 
 
 def connect(application, http_server, worker, prefix=""):
@@ -26,7 +25,6 @@ def connect(application, http_server, worker, prefix=""):
 applications = {
     "/status": status_doc,
     "/counters": counters_doc,
-    "/crossfilter": crossfilter_doc,
     "/system": systemmonitor_doc,
     "/profile": profile_doc,
     "/profile-server": profile_server_doc,
