@@ -1,3 +1,5 @@
+from __future__ import annotations
+
 import logging
 import math
 import warnings
@@ -7,12 +9,13 @@ import toolz
 from dask.system import CPU_COUNT
 from dask.widgets import get_template
 
-from ..nanny import Nanny
-from ..scheduler import Scheduler
-from ..security import Security
-from ..worker import Worker, parse_memory_limit
-from .spec import SpecCluster
-from .utils import nprocesses_nthreads
+from distributed.deploy.spec import SpecCluster
+from distributed.deploy.utils import nprocesses_nthreads
+from distributed.nanny import Nanny
+from distributed.scheduler import Scheduler
+from distributed.security import Security
+from distributed.worker import Worker
+from distributed.worker_memory import parse_memory_limit
 
 logger = logging.getLogger(__name__)
 
@@ -56,8 +59,8 @@ class LocalCluster(SpecCluster):
         Set to True if using this cluster within async/await functions or within
         Tornado gen.coroutines.  This should remain False for normal use.
     blocked_handlers: List[str]
-        A list of strings specifying a blacklist of handlers to disallow on the Scheduler,
-        like ``['feed', 'run_function']``
+        A list of strings specifying a blocklist of handlers to disallow on the
+        Scheduler, like ``['feed', 'run_function']``
     service_kwargs: Dict[str, Dict]
         Extra keywords to hand to the running services
     security : Security or bool, optional
